@@ -542,6 +542,16 @@ RPN(rpn_cast)
   pop(2);
 }
 
+RPN(rpn_prec)
+{
+  genprec(d->fout, preclevel);
+}
+
+RPN(rpn_bitprec)
+{
+  genbitprec(d->fout, preclevel);
+}
+
 RPN(rpn_code) { push(1); LVL(0)=-1; rpn_cast(d,stk); }
 
 RPN(rpn_wrapper)
@@ -575,7 +585,8 @@ RPN(rpn_cookie)
     gpfunc *gp = lfunc + findfunction(entryname(arg[0]));
     if (gp->spec==GPuser && gp->user->wrapper>=0)
     {
-      if (funcmode(*gp)&(1<<Mprec))
+      gpfunc *wr = lfunc + gp->user->wrapper;
+      if ((funcmode(*gp)&(1<<Mprec)) && !(funcmode(*wr)&(1<<Mprec)))
         genfuncbydesc(d->fout, nb-1,arg+1,FC_tovecprec,d->nerr);
       else
       {
@@ -651,6 +662,7 @@ const char *gencoderpn(FILE *fout, const char *p, gpdescarg *rule, int nbarg, in
     {"parens",0,rpn_parens},{"stdref",0,rpn_stdref},
     {"str_format",1,rpn_str_format},{"str_raw",1,rpn_str_raw},
     {"code",1,rpn_code},{"cast",2,rpn_cast},
+    {"prec",0,rpn_prec},{"bitprec",0,rpn_bitprec},
     {"format_string",1,rpn_format_string}, {"format_args",1,rpn_format_args},
     {"cookie",1,rpn_cookie}, {"wrapper",1,rpn_wrapper},
     {NULL,0,NULL}
