@@ -102,7 +102,10 @@ int guesstype(int n)
       return Gstr;
     }
   case Fentry:
-    return vartype(ctxstack[getvarerr(n)]);
+    {
+      long v = getvar(n);
+      return v<0 ? Ggen: vartype(ctxstack[v]);
+    }
   case Ffunction:
     if (!strcmp(entryname(n),"length"))
       return Glg;
@@ -133,17 +136,30 @@ void genequaltosmall(int binf,int bsup, int *tv, int def)
 
 int gengerepile(int bl)
 {
-  int arg[4];
+  int arg[3];
   block[bl].gc|=(1<<GCneeded);
   arg[0]=newnode(Fentry,newentry("btop"),-1);
   pushvar(arg[0],0,Gpari_sp,newcall("_avma",-1));
-  arg[1]=newnode(Fentry,newentry("st_lim"),-1);
-  pushvar(arg[1],0,Gpari_sp,
-      newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
-  arg[2]=newsmall(bl);
+  arg[1]=newsmall(bl);
+  if (FC_gc_needed>=0)
+    arg[2]=GNOARG;
+  else
+  {
+    arg[2]=newnode(Fentry,newentry("st_lim"),-1);
+    pushvar(arg[2],0,Gpari_sp,
+        newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
+  }
   return stacktoargs(arg,3);
 }
 
+int gengerepilereturn(int bl)
+{
+  int arg[2];
+  arg[0]=newnode(Fentry,newentry("ltop"),-1);
+  arg[1]=newsmall(bl);
+  return stacktoargs(arg,2);
+}
+
 int newgetg(int v, int l, const char *t)
 {
   return newcall("_cgetg", newnode(Flistarg, newnode(Flistarg,
@@ -153,8 +169,10 @@ int newgetg(int v, int l, const char *t)
 void makesubblock_ctx(int n, int savx)
 {
   int bl3=newblock();
+  int level = preclevel;
   genblock(n,-1);
   makeblock(bl3,n,newleaf(n),-1,savx);
+  preclevel = level;
 }
 
 void makesubblock(int n)
@@ -258,8 +276,11 @@ geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
       break;
     case PPdefault:
     case PPdefaultmulti:
-      if (i<nb) tree[arg[i]].t = guesstype(arg[i]);
-      argo[k++] = i<nb ? arg[i] : GNOARG;
+      if (i < nb)
+      {
+        tree[arg[i]].t = guesstype(arg[i]);
+        argo[k++] = arg[i];
+      }
       break;
     case PPstar:
       i=nb-1;
@@ -307,9 +328,11 @@ genblockfunciter(int n, gpfunc *gp)
     gpm=gengerepile(bl2);
   if (tv==Gnotype)
   {
+    gpdescarg *dv;
     tree[vprime].t = iter->type;
-    gpdescarg *da=descfindrules(1, &vprime, lfunc+iter->next);
-    tv = da->type;
+    dv=descfindrules(1, &vprime, lfunc+iter->next);
+    if (!dv) die(n, "no suitable iterator_next found");
+    tv = dv->type;
   }
   pushvar(var,1<<Cuser|1<<Ccompo,tv,-1);
   lab=makesubblock_label(seq,Lbreak);
@@ -317,7 +340,7 @@ genblockfunciter(int n, gpfunc *gp)
     seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
   if (autogc)
     seq=addseqright(seq,newcall("_gerepilemany",gpm));
-  cseq=newnode(Fassign,var,newcall(next,vprime));
+  cseq=newnode1(Fassign,var,newcall(next,vprime));
   tree[cseq].m=Mparens;
   aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,cseq,seq)));
   if (lab.go)
@@ -419,8 +442,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         arg[0] = newleaf(pred);
       }
       genblock(arg[0],n);
-      arg[1]=geninsertvar(arg[1],ret);
-      makesubblock(arg[1]);
+      if (arg[1]!=GNOARG)
+      {
+        arg[1]=geninsertvar(arg[1],ret);
+        makesubblock(arg[1]);
+      }
       if(nb==3)
       {
         arg[2]=geninsertvar(arg[2],ret==-1?-1:newleaf(ret));
@@ -490,7 +516,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       }
       if (tree[arg[1]].f==Ftag)
         functype(*gp)=tree[arg[1]].y;
-      for(;*p;p++) if (*p=='p') funcmode(*gp)|=(1<<Mprec);
+      for(;*p;p++) if (*p=='p' || *p=='b') funcmode(*gp)|=(1<<Mprec);
       gp->spec=GPinstalled;
     }
     break;
@@ -515,7 +541,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Fassign,var,vinf);
+    arg[0]=newnode1(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
@@ -550,7 +576,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bl2=newblock();
       if (autogc)
         gpm=gengerepile(bl2);
-      next = newnode(Fassign,newleaf(var),newnode(Ftag,
+      next = newnode1(Fassign,newleaf(var),newnode(Ftag,
             newcall("_call_GG",newnode(Flistarg,newnode(Flistarg,
                   newleaf(vfvnext),newleaf(vfvdat)),newleaf(var))),Gvec));
       lab=makesubblock_label(seq,Lbreak);
@@ -568,15 +594,19 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,sav0,savx);
     }
     break;
+  case GPfordivfactored:
   case GPfordiv:
     {
       int vdiv,div;
       int vloop;
+      int factored = gp->spec == GPfordivfactored;
+      int vt = factored ? Gvec: Gint;
       nb=genlistargs(n,arg,3,3);
       seq=arg[2];
       checkisvar(arg[1],gp->gpname,&var,&tv);
       newdecl(1<<Cconst,Gvec,-1,&vdiv);
-      div=newcall("divisors",arg[0]);
+      div=newcall("divisors", factored ?
+          newnode(Flistarg, arg[0], newsmall(1)) : arg[0]);
       bseq=addseqright(bseq,geninsertvar(div,vdiv));
       genblock(bseq,-1);
       savx=s_ctx.n;
@@ -585,7 +615,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (autogc)
         gpm=gengerepile(bl2);
       newdecl(0,Gsmall,-1,&vloop);
-      if (tv==Gnotype) tv=Gint;
+      if (tv==Gnotype) tv = vt;
       pushvar(var,1<<Cuser,tv,-1);
       lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
@@ -593,10 +623,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Fassign,newleaf(var),
+          newnode1(Fassign,newleaf(var),
             newnode(Ftag,
-              newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Fassign,vloop,newsmall(1));
+              newcoeff(newleaf(vdiv),newleaf(vloop),-1), vt)),seq);
+      arg[0]=newnode1(Fassign,vloop,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
@@ -627,18 +657,23 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&vindex);
         newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&lstep);
         bseq=addseqright(bseq,geninsertvar(newcall("length",newleaf(vstep)),newleaf(lstep)));
-        sarg[0] = newnode(Fassign,newleaf(vindex),newsmall(1));
+        sarg[0] = newnode1(Fassign,newleaf(vindex),newsmall(1));
         sarg[1] = newleaf(lstep);
         sarg[2] = newnode(Fmatcoeff,newleaf(vstep),
-            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNOARG),-1));
+            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNORANGE),-1));
         sstep = newcall("sum", stacktoargs(sarg,3));
+        if (tv==Gnotype) tv=Ggen;
       }
       else
       {
         isvec=0;
         dosign=tree[bstep].f!=Fsmall;
         if (dosign)
+        {
           sstep = newleaf(vstep);
+          if (tv==Gnotype) tv=Ggen;
+        } else
+          genequaltosmall(binf,bsup,&tv,Ggen);
       }
       if (dosign)
       {
@@ -652,15 +687,13 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bl2=newblock();
       if (autogc)
         gpm=gengerepile(bl2);
-      if (tv==Gnotype)
-        tv=Ggen;
       pushvar(var,1<<Cuser,tv,-1);
       lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      arg[0]=newnode(Fassign,var,vinf);
+      arg[0]=newnode1(Fassign,var,vinf);
       if ( !dosign )
       {
         if ( tree[bstep].x>0 )
@@ -742,7 +775,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (bsup==GNOARG) nbs=2; else args[2]=vsup;
       aseq=addseqleft(newcall((tv==Gsmall?"_u_forprime_init":"_forprime_init"),
             stacktoargs(args,nbs)),aseq);
-      bseq=newnode(Fassign,var,newcall(tv==Gsmall?"_u_forprime_next":"_forprime_next_",vprime));
+      bseq=newnode1(Fassign,var,newcall(tv==Gsmall?"_u_forprime_next":"_forprime_next_",vprime));
       tree[bseq].m=Mparens;
       aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,bseq,seq)));
     }
@@ -799,7 +832,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     makesubblock_label(seq,Lnobrk);
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Fassign,var,vinf);
+    arg[0]=newnode1(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
@@ -832,10 +865,10 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Fassign,newleaf(var),
+          newnode1(Fassign,newleaf(var),
             newnode(Ftag,
               newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Fassign,vloop,newsmall(1));
+      arg[0]=newnode1(Fassign,vloop,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
@@ -849,17 +882,22 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl1=newblock();
     savx=s_ctx.n;
     seq=arg[1];
-    if(autogc)
-      gpm=gengerepile(bl1);
-    genblock_label(arg[0],n,Lnobrk);
-    lab=makesubblock_label(seq,Lbreak);
-    if (lab.ne)
-      seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
-    if (autogc)
-      seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    if (seq != GNOARG)
+    {
+      if(autogc)
+        gpm=gengerepile(bl1);
+      genblock_label(arg[0],n,Lnobrk);
+      lab=makesubblock_label(seq,Lbreak);
+      if (lab.ne)
+        seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+      if (autogc)
+        seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    }
+    else
+      genblock(arg[0],n);
     arg[1]=seq;
     aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb));
-    if (lab.go)
+    if (seq != GNOARG && lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblock(bl1,n,aseq,ret,savx);
     break;
@@ -884,7 +922,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl(0,Gsmall,-1,&var);
     aseq=addseqright(aseq,newgetg(ret,newleaf(vsup),
           gp->spec==GPvector?"t_VEC":gp->spec==GPvectorv?"t_COL":"t_VECSMALL"));
-    arg[0]=newnode(Fassign,var,newsmall(1));
+    arg[0]=newnode1(Fassign,var,newsmall(1));
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=newcoeff(newleaf(ret),newleaf(var),-1);
     arg[2]=geninsertvar(seq,arg[2]);
@@ -928,7 +966,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         seq=arg[4];
       aseq=newgetg(newcoeff(newleaf(ret),newleaf(var1),-1),
           newleaf(vsup2),"t_COL");
-      arg[0]=newnode(Fassign,var2,newsmall(1));
+      arg[0]=newnode1(Fassign,var2,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var2),vsup2);
       arg[3]=newopcall(OPpp,newleaf(var2),-1);
       arg[2]=newcoeff(newleaf(ret),newleaf(var2),newleaf(var1));
@@ -936,7 +974,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       makesubblock_label(arg[2],Lnobrk);
       arg[2]=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
       aseq=newgetg(ret,newleaf(vsup1),"t_MAT");
-      arg[0]=newnode(Fassign,var1,newsmall(1));
+      arg[0]=newnode1(Fassign,var1,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var1),vsup1);
       arg[3]=newopcall(OPpp,newleaf(var1),-1);
       aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
@@ -957,6 +995,37 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (tree[p].f==Ftag) tree[p]=tree[n];
     }
     break;
+  case GPself:
+    {
+      int str = newstringnode(lfunc[currfunc].gpname,-1);
+      int aseq = newcall("_strtoclosure",str);
+      tree[n] = tree[aseq];
+    }
+    break;
+  case GPlocalprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"prec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cauto), Gsmall, -1);
+      y = newnode1(Fassign, vn, newcall("_ndec2prec", y));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
+  case GPlocalbitprec:
+    {
+      char buf[64];
+      int vn;
+      sprintf(buf,"bitprec%d",++preclevel);
+      vn = newnode(Fentry, newentry(strdup(buf)), -1);
+      pushvar(vn, (1<<Cauto), Gsmall, -1);
+      y = newnode1(Fassign, vn, newnode(Ftag, newleaf(y), Gsmall));
+      genblock(y, n);
+      tree[n].y = y;
+    }
+    break;
   case GPreturn:
     {
       int iferr_reset=0;
@@ -994,10 +1063,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       block[bl2].egc=var;
       if (autogc)
       {
-        arg[0]=newnode(Fentry,newentry("ltop"),-1);
-        arg[1]=GNOARG;
-        arg[2]=newsmall(bl2);
-        gpm=stacktoargs(arg,3);
+        gpm=gengerepilereturn(bl2);
         aseq=addseqright(aseq,newcall("_gerepilemany",gpm));
       }
       if (iferr_reset)
@@ -1201,6 +1267,10 @@ int gentypefuncspec(int n, gpfunc *gp)
     gentype(y);
     tree[n].m=(tree[y].m&MODHERIT)|(1<<Msidef);
     return Gvec;
+  case GPmakevecsmall:
+    gentype(y);
+    tree[n].m=(tree[y].m&MODHERIT)|(1<<Msidef);
+    return Gvecsmall;
   case GPclosure:
     nb=genlistargs(n,arg,1,STACKSZ);
     gentypemode(nb,arg,&tree[n].m);
@@ -1319,8 +1389,13 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       {
         genindentline(fout,"while (");
         gencast(fout,arg[0],Gbool);
-        fprintf(fout,")\n");
-        genbrace(fout,arg[1]);
+        if (arg[1]==GNOARG)
+          fprintf(fout,");\n");
+        else
+        {
+          fprintf(fout,")\n");
+          genbrace(fout,arg[1]);
+        }
       }
       else
       {
@@ -1359,9 +1434,14 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       nb=genlistargs(n,arg,2,2);
       if ( tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
       {
-        genindentline(fout,"do\n");
-        genbrace(fout,arg[1]);
-        fprintf(fout," while(");/*KB style*/
+        if (arg[1]!=GNOARG)
+        {
+          genindentline(fout,"do\n");
+          genbrace(fout,arg[1]);
+          fprintf(fout," while(");/*KB style*/
+        }
+        else
+          genindentline(fout,"while(");
         gencast(fout,arg[0],Gnegbool);
         fprintf(fout,");\n");
       }
@@ -1474,6 +1554,7 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     fprintf(fout,")\n");
     genbrace(fout,arg[1]);
     break;
+  case GPmakevecsmall:
   case GPmakevec:
     nb=genlistargs(n,arg,1,STACKSZ);
     genindent(fout);
@@ -1487,8 +1568,8 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     }
     break;
   case GPgpmany:
-    nb=genlistargs(n,arg,3,STACKSZ);
-    bl=block+tree[arg[2]].x;
+    nb=genlistargs(n,arg,2,3);
+    bl=block+tree[arg[1]].x;
     if (!(bl->gc&(1<<GCneeded)))
       break;
     genindent(fout);
@@ -1513,11 +1594,19 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     {
       int parg[STACKSZ+1];
       /*FIXME:Putting an if statement in the wilderness is not secure*/
-      if (arg[1]!=GNOARG)
+      if (nb==3)
       {
-        parg[0]=arg[1]; parg[1]=arg[0];
         fprintf(fout,"if (");
-        genfuncbydesc(fout,2,parg,FC_low_stack_lim,n);
+        if (arg[2]==GNOARG)
+        {
+          parg[0]=arg[0];
+          genfuncbydesc(fout,1,parg,FC_gc_needed,n);
+        }
+        else
+        {
+          parg[0]=arg[2]; parg[1]=arg[0];
+          genfuncbydesc(fout,2,parg,FC_low_stack_lim,n);
+        }
         fprintf(fout,")\n");
         indent++;
         genindent(fout);
@@ -1530,6 +1619,11 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       fprintf(fout,";\n");
     }
     break;
+  case GPlocalprec:
+  case GPlocalbitprec:
+    gencode(fout, y);
+    preclevel++;
+    break;
   case GPreturn:
     fprintf(fout,"return");
     if (functype(lfunc[currfunc])!=Gvoid)
