@@ -22,7 +22,7 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 static int toplevel=0;
 int initnode=0;
 
-const char *basename(const char *s)
+static const char *basename(const char *s)
 {
   long i;
   long l=strlen(s);
@@ -109,22 +109,24 @@ void gentoplevel(int n)
   case Fseq:
     if (x>=0 && tree[x].f!=Fdeffunc && tree[x].f!=Fseq)
     {
-      if (isfunc(x,"global"))
-        genpredeclaration(tree[x].y,x);
-      if (isfunc(x,"local"))
-        warning(x,"`local' at top-level discarded");
+      int dx = detag(x);
+      if (isfunc(dx,"global"))
+        genpredeclaration(tree[dx].y,dx);
+      if (isfunc(dx,"local"))
+        warning(dx,"`local' at top-level discarded");
       else
         toplevel=addseqright(toplevel,x);
       tree[n].x=GNIL;
     }
     else
       gentoplevel(x);
-    if (y>=0 && tree[y].f!=Fdeffunc)
+    if (y>=0 && tree[y].f!=Fdeffunc && tree[y].f!=Fseq)
     {
-      if (isfunc(y,"global"))
-        genpredeclaration(tree[y].y,y);
-      if (isfunc(y,"local"))
-        warning(y,"`local' at top-level discarded");
+      int dy = detag(y);
+      if (isfunc(dy,"global"))
+        genpredeclaration(tree[dy].y,dy);
+      if (isfunc(dy,"local"))
+        warning(dy,"`local' at top-level discarded");
       else
         toplevel=addseqright(toplevel,y);
       tree[n].y=GNIL;
